/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * fetch-generic.cpp - fetches http:// and ftp:// urls using libcurl
 *
 * 2005-01-19	dropped old fetch-http, added this libcurl wrappz0r
 */

#include "fetch-generic.h"
#include <unistd.h>

extern "C"
{
	mwcollectd::Module * __attribute__ ((dllexport)) CreateInstance(mwcollectd::MalwareCollector * pCollector)
	{
		return new FetchGenericModule(pCollector);
	}
}


FetchGenericModule::FetchGenericModule(MalwareCollector * pCollector)
{
	m_pCollector = pCollector;
}

FetchGenericModule::~FetchGenericModule()
{
	free(m_szConfigPath);
}


void FetchGenericModule::start()
{
	{
		if(!(m_pCurlStack = curl_multi_init()))
			m_pLogManager->log(LL_CRITICAL, "Could not create CURL Stack!");

		m_nQueued = 0;
	}

	m_pCollector->getFetchDispatcher()->addProtocolHandler(this);
}

void FetchGenericModule::loop()
{
	int iQueue;

	while(curl_multi_perform(m_pCurlStack, &iQueue) == CURLM_CALL_MULTI_PERFORM);

	if(m_nQueued > iQueue)
	{
		CURLMsg * pMessage;

		while((pMessage = curl_multi_info_read(m_pCurlStack, &iQueue)))
		{
			if(pMessage->msg == CURLMSG_DONE)
			{
				fetch_generic_info_t * pInfo;
				char * szUrl;

				curl_easy_getinfo(pMessage->easy_handle, CURLINFO_PRIVATE, (char * *) &pInfo);
				fclose(pInfo->pHandle);				

				if(pMessage->data.result)
				{					
					m_pLogManager->log(LL_INFO, "Downloading of %s failed with error code %u.", pInfo->szUrl, pMessage->data.result);

					unlink(pInfo->szFileName);
				}
				else
				{
					curl_easy_getinfo(pMessage->easy_handle, CURLINFO_EFFECTIVE_URL, &szUrl);
					m_pLogManager->log(LL_INFO, "Downloading of %s (%s) successful.", pInfo->szUrl, szUrl);

					m_pCollector->getSubmissionDispatcher()->submitFile(pInfo->szFileName);
				}

				curl_multi_remove_handle(m_pCurlStack, pMessage->easy_handle);

				free(pInfo->szFileName);
				free(pInfo->szUrl);
				delete pInfo;

				if(!--m_nQueued)
					m_pCollector->getNetworkInterface()->fastpollHack(false);
			}
		}
	}
}

void FetchGenericModule::stop()
{
	curl_multi_cleanup(m_pCurlStack);
}


bool FetchGenericModule::supportProtocol(const char * szProtocol)
{
	return ((strcmp(szProtocol, "http") == 0) || (strcmp(szProtocol, "ftp") == 0));
}

void FetchGenericModule::fetchFile(const char * szUrl)
{
	CURL * pHandle = curl_easy_init();
	char * szPath, * szLocalFile;
	const char * szUserAgent;
	long lMaxFilesize, lRedirs;
	fetch_generic_info_t * pInfo = new fetch_generic_info_t;
	FILE * pFile;

	if(!pHandle)
	{
		m_pLogManager->log(LL_CRITICAL, "curl_easy_init() failed!");

		return;
	}

	{
		asprintf(&szPath, "%s:temp-directory", m_szConfigPath);
		asprintf(&szLocalFile, "%smwc-fetchgen.%u.%08x", m_pConfiguration->getString(szPath, "/tmp/"), time(0), rand());
		free(szPath);

		asprintf(&szPath, "%s:user-agent", m_szConfigPath);
		szUserAgent = m_pConfiguration->getString(szPath, "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)");
		free(szPath);

		asprintf(&szPath, "%s:max-filesize", m_szConfigPath);
		lMaxFilesize = m_pConfiguration->getLong(szPath, 10485760);
		free(szPath);

		asprintf(&szPath, "%s:redirects", m_szConfigPath);
		lRedirs = m_pConfiguration->getLong(szPath, 3);
		free(szPath);
	}

	{
		if(!(pFile = fopen(szLocalFile, "wb")))
		{
			free(szLocalFile);

			return;
		}

		pInfo->szFileName = szLocalFile;
		pInfo->pHandle = pFile;
		pInfo->szUrl = strdup(szUrl);
	}

	{
		curl_easy_setopt(pHandle, CURLOPT_NOPROGRESS, 1);
		curl_easy_setopt(pHandle, CURLOPT_NOSIGNAL, 1);

		curl_easy_setopt(pHandle, CURLOPT_URL, pInfo->szUrl);

		curl_easy_setopt(pHandle, CURLOPT_WRITEFUNCTION, FetchGenericModule::WriteCallback);
		curl_easy_setopt(pHandle, CURLOPT_WRITEDATA, pFile);
		curl_easy_setopt(pHandle, CURLOPT_PRIVATE, (char *) pInfo);

		if(strncmp(szUrl, "http://", 7) == 0)
		{
			curl_easy_setopt(pHandle, CURLOPT_FOLLOWLOCATION, (lRedirs > 0 ? 1 : 0));
			curl_easy_setopt(pHandle, CURLOPT_MAXREDIRS, lRedirs);
			curl_easy_setopt(pHandle, CURLOPT_USERAGENT, szUserAgent);
		}
		else if(strncmp(szUrl, "ftp://", 6) == 0)
		{
			curl_easy_setopt(pHandle, CURLOPT_FTP_SSL, CURLFTPSSL_NONE);
		}

		curl_easy_setopt(pHandle, CURLOPT_SSL_VERIFYHOST, false);
		curl_easy_setopt(pHandle, CURLOPT_SSL_VERIFYPEER, false);		

		curl_easy_setopt(pHandle, CURLOPT_MAXFILESIZE, lMaxFilesize);
		curl_easy_setopt(pHandle, CURLOPT_FORBID_REUSE, 1);
	}

	curl_multi_add_handle(m_pCurlStack, pHandle);
	++ m_nQueued;

	m_pCollector->getNetworkInterface()->fastpollHack(true);
}


size_t FetchGenericModule::WriteCallback(void * pData, size_t nBlocksize, size_t nBlocks, void * pUserData)
{
	return fwrite(pData, 1, nBlocksize * nBlocks, (FILE *) pUserData);
}
