/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * ftftp.cpp - the tftp fetching module header with coordinator and slaves implementation
 *
 * 2005-02-13	 you'll all die.
 */

#include "ftftp.h"
#include <time.h>

using namespace mwcollectd;

extern "C"
{
  Module * __attribute__ ((dllexport)) CreateInstance(MalwareCollector * pCollector)
  {
    return ((Module *) (new TftpCoordinator(pCollector)));
  }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void TftpCoordinator::start()
{
	m_pCollector->getFetchDispatcher()->addProtocolHandler(this);

	{
		char * szPath;

		asprintf(&szPath, "%s:timeout", m_szConfigPath);
		m_ulTimeout =  (unsigned long) m_pConfig->getLong(szPath, 30);
		free(szPath);

		asprintf(&szPath, "%s:retries", m_szConfigPath);
		m_ulRetries =  (unsigned long) m_pConfig->getLong(szPath, 16);
		free(szPath);

		asprintf(&szPath, "%s:block-strict", m_szConfigPath);
		m_bBlockStrict = m_pConfig->getLong(szPath, 0) != 0;
		free(szPath);
	}
}

void TftpCoordinator::loop()
{
	for(std::list<TftpFetchSession *>::iterator i = m_lpSessions.begin(); i != m_lpSessions.end(); ++i)
		(* i)->checkTimeout();
}

void TftpCoordinator::stop()
{
	m_pCollector->getFetchDispatcher()->removeProtocolHandler(this);
}

bool TftpCoordinator::supportProtocol(const char * szProtocol)
{
	return (strcmp(szProtocol, "tftp") == 0);
}

void TftpCoordinator::fetchFile(const char * szURL)
{
	char szFile[256], szHost[256];

	{
		const char * szWalk = strstr(szURL, "://");
		unsigned int i;

		if(!szWalk)
			return;

		szWalk += 3;

		for(i = 0; * szWalk && * szWalk != '/'; ++i, ++szWalk)
			szHost[i] = * szWalk;

		szHost[i] = 0;
		++szWalk;

		strncpy(szFile, szWalk, sizeof(szFile));
	}

	{
		TftpFetchSession * pSession = new TftpFetchSession(this, m_pCollector, m_pLogManager, m_ulTimeout, m_bBlockStrict, m_ulRetries); 
		int iSocket = m_pCollector->getNetworkInterface()->createUnboundDatagramSocket(TftpCoordinator::incomingNotifier, TftpCoordinator::terminationNotifier, pSession);

		if(iSocket < 0)
			m_pLogManager->log(LL_CRITICAL, "Could not create UDP socket to fetch %s!", szURL);
		else
			pSession->fetchFile(iSocket, m_pCollector->getNetworkInterface()->lookupHost(szHost), szFile);

		m_lpSessions.push_back(pSession);
	}
}


void TftpCoordinator::terminationNotifier(bool bServer, void * pUserData)
{
	((TftpFetchSession *) pUserData)->removeMe();

	delete ((TftpFetchSession *) pUserData);
}

void TftpCoordinator::incomingNotifier(int iHandle, unsigned char * pucData, unsigned int nLength, void * pUserData, unsigned long ulAddress, unsigned short usPort)
{
	((TftpFetchSession *) pUserData)->incomingData(iHandle, pucData, nLength, ulAddress, usPort);
}

void TftpCoordinator::removeChild(TftpFetchSession * pSession)
{
	for(std::list<TftpFetchSession *>::iterator i = m_lpSessions.begin(); i != m_lpSessions.end(); ++i)
	{
		if (pSession == (*i))
		{
			m_pLogManager->log(LL_SPAM, "%s Removing Child %x", __PRETTY_FUNCTION__, pSession);

			m_lpSessions.erase(i);
			return;
		}
	}	
}



// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TftpFetchSession::~TftpFetchSession()
{

	if(m_szFileName)
		free(m_szFileName);

	if(m_pucBuffer)
		free(m_pucBuffer);

	if(m_pucLastSend)
		free(m_pucLastSend);
}

void TftpFetchSession::removeMe()
{
	m_pFather->removeChild(this);
}

void TftpFetchSession::fetchFile(int iHandle, unsigned long ulHost, const char * szFileName)
{
	{
		m_iHandle = iHandle;
		m_szFileName = strdup(szFileName);
		m_ulAddress = ulHost;
	}

	{		
		unsigned char * pucRequest = (unsigned char *) malloc(strlen(szFileName) + 9);
		m_pucLastSend = (unsigned char *) malloc(strlen(szFileName) + 9);

		* ((unsigned short *) pucRequest) = 0x0100; // read request
		strcpy((char *) pucRequest + 2, szFileName);
		strcpy((char *) pucRequest + strlen(szFileName) + 3, "octet");

		m_pCollector->getNetworkInterface()->sendDatagram(iHandle, pucRequest, strlen(szFileName) + 9, ulHost, 69);

		m_ulNumTimeout = 0;
		m_ulLastSend = strlen(szFileName) + 9;			
		memcpy(m_pucLastSend, pucRequest, m_ulLastSend);
		m_usPort = 69;

		free(pucRequest);
	}

	m_ulLastEvent = time(0);
	m_ulReceived = 0;
	m_usBlock = 0;
}

void TftpFetchSession::checkTimeout()
{
	if(time(0) - m_ulLastEvent >= m_ulTimeout)
	{
		if(m_ulNumTimeout < m_ulMaxTimeout)
		{
			 m_pCollector->getNetworkInterface()->sendDatagram(m_iHandle, m_pucLastSend, m_ulLastSend, m_ulAddress, m_usPort);

			 m_ulLastEvent = time(0);
		}
		else
		{
			m_pLogManager->log(LL_INFO, "Receive of %s timed out awaiting block #%u (%u/%u tries).", m_szFileName, (unsigned int) m_usBlock, m_ulNumTimeout, m_ulMaxTimeout);

			// removeMe();
			m_pCollector->getNetworkInterface()->closeSocket(m_iHandle);
		}

		++ m_ulNumTimeout;
	}
}

void TftpFetchSession::incomingData(int iHandle, unsigned char * pucData, unsigned int nLength, unsigned long ulSource, unsigned short usSource)
{
	unsigned char pucAck[4];
	unsigned short usBlock;

	if(ulSource != m_ulAddress)
		return;

	if(nLength < 4)
		return;

	if(* ((unsigned short *) pucData) != 0x0300) // no data packet
	{
		if(* ((unsigned short *) pucData) == 0x0500)
		{
			unsigned short usCode = ntohs(* ((unsigned short *) (pucData + 2)));

			if(nLength > 4 && !pucData[nLength - 1])
				m_pLogManager->log(LL_INFO, "Error fetching %s, server reported error code %u: \"%s\".", m_szFileName, usCode, pucData + 4);				
			else
				m_pLogManager->log(LL_CRITICAL, "Error fetching %s, server reported error code %u (unterminated error message)!");

			m_pCollector->getNetworkInterface()->closeSocket(iHandle);
		}

		return;
	}

	usBlock = ntohs(* ((unsigned short *) (pucData + 2)));

	m_ulLastEvent = time(0);

	if(m_bBlockStrict && usBlock != m_usBlock + 1) // drop unordered packets as they have to be retransmitted
	{
		m_pLogManager->log(LL_SPAM, "TFTP: dropping packet with block number %u <- out of order", (unsigned int) m_usBlock);

		return;
	}

	++m_usBlock;

	* ((unsigned short *) pucAck) = 0x0400;
	* ((unsigned short *) (pucAck + 2)) = htons(usBlock);
	m_pCollector->getNetworkInterface()->sendDatagram(iHandle, pucAck, 4, ulSource, usSource);

	memcpy(m_pucLastSend, pucAck, 4); // buffer is still filename + 9 bytes long, should be enough (+ 9 [!!!]) :D
	m_ulLastSend = 4;
	m_usPort = usSource;
	
	if(nLength > 4)
	{
		m_pucBuffer = (unsigned char *) realloc(m_pucBuffer, m_ulReceived + nLength - 4);
		memcpy(m_pucBuffer + m_ulReceived, pucData + 4, nLength - 4);
		m_ulReceived += nLength - 4;
		m_ulNumTimeout = 0;
	}

	if(nLength != 516) // last packet
	{
		m_pLogManager->log(LL_INFO, "Finished fetching %s.", m_szFileName);
		m_pCollector->getSubmissionDispatcher()->submitData(m_pucBuffer, m_ulReceived);

		m_pCollector->getNetworkInterface()->closeSocket(iHandle);
	}
}
