/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * ftftp.h - the tftp fetching module header with coordinator and slaves
 *
 * 2005-02-13	 you'll all die.
 */

#ifndef __FETCH_TFTP_H
#define __FETCH_TFTP_H

#include <mwcollectd.h>
#include <list>

using namespace mwcollectd;

class TftpCoordinator;

class TftpFetchSession
{
public:
	TftpFetchSession(TftpCoordinator * pFather, MalwareCollector * pCollector, LogManager * pLogManager, unsigned long ulTimeout, bool bBlockStrict, unsigned long ulMaxTimeout)
	{ m_pFather = pFather; m_pCollector = pCollector; m_pLogManager = pLogManager; m_ulTimeout = ulTimeout; m_szFileName = 0; m_pucBuffer = 0; m_bBlockStrict = bBlockStrict; m_pucLastSend = 0; m_ulLastSend = 0; m_ulMaxTimeout = ulMaxTimeout; }

	~TftpFetchSession();

	void fetchFile(int iHandle, unsigned long ulHost, const char * szFileName);
	void incomingData(int iHandle, unsigned char * pucData, unsigned int nLength, unsigned long ulSource, unsigned short usSource);
	void checkTimeout();
	void removeMe();

protected:
	TftpCoordinator * m_pFather;
	MalwareCollector * m_pCollector;
	LogManager * m_pLogManager;
	unsigned long m_ulTimeout;
	
	int m_iHandle;
	char * m_szFileName;

	unsigned long m_ulAddress;
	unsigned char * m_pucBuffer;
	unsigned long m_ulReceived;
	unsigned long m_ulLastEvent;
	unsigned short m_usBlock;
	bool m_bBlockStrict;
	
	unsigned char * m_pucLastSend;
	unsigned int  m_ulLastSend;	
	unsigned long m_ulNumTimeout;
	unsigned long m_ulMaxTimeout;
	unsigned short m_usPort;
};

class TftpCoordinator: public Module, ProtocolHandler
{
public:
	TftpCoordinator(MalwareCollector * pCollector) { m_pCollector = pCollector; }
	virtual ~TftpCoordinator() { }
	virtual void setLogManager(LogManager * pLogManager) { m_pLogManager = pLogManager; }
	virtual void setConfiguration(Configuration * pConfig, const char * szPath) { m_pConfig = pConfig; m_szConfigPath = strdup(szPath); }

	virtual void start();
	virtual void loop();
	virtual void stop();

	virtual void fetchFile(const char * szUrl);
	virtual bool supportProtocol(const char * szProtocol);

	static void incomingNotifier(int iHandle, unsigned char * pucPacket, unsigned int nLength, void * pUserData, unsigned long ulSource, unsigned short usSource);
	static void terminationNotifier(bool fServer, void * pUserData);

	void removeChild(TftpFetchSession * pSession);

protected:
	MalwareCollector * m_pCollector;
	LogManager * m_pLogManager;
	Configuration * m_pConfig;
	char * m_szConfigPath;

	unsigned long m_ulTimeout;
	unsigned long m_ulRetries;
	bool m_bBlockStrict;

	std::list<TftpFetchSession *> m_lpSessions, m_lpRemove;
};

#endif // __FETCH_TFTP_H
