/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * log.h - definition of basic FileLogger, LogManager and the LogLevels
 *
 * 2005-01-05	 invented new design, double checked, seems to be ok
 *
 */

#ifndef __MWCOLLECTD_LOG_H
#define __MWCOLLECTD_LOG_H

#include <list>
#include <stdio.h>

namespace mwcollectd
{
	typedef enum
	{
		LL_NONE = 0,
		LL_SPAM,
		LL_DEBUG,
		LL_INFO,
		LL_CRITICAL,
	} log_level_t;

	class BasicLogger
	{
	public:
		virtual void log(log_level_t llLevel, const char * szMessage) = 0;
		virtual void setLogLevel(log_level_t llFilter) = 0;
	};

	class LogManager
	{
	public:
		LogManager(log_level_t llDefaultFilter = LL_NONE);
		virtual ~LogManager();

		virtual void addLogger(BasicLogger * pLogger);
		virtual void removeLogger(BasicLogger * pLogger);
		virtual std::list<BasicLogger *> * loggerList();

		virtual void log(log_level_t llLevel, const char * szFormat, ...);


		static log_level_t LogManager::mapLogLevel(const char * szInput)
		{ // maps a string to the corespondending LL_* constant
			static char * pszLevels[] = { "none", "spam", "debug", "info", "critical", 0 };
			char * * pszWalk = pszLevels;
			log_level_t llIterate = LL_NONE;

			while(* pszWalk)
			{
				if(!strcmp(* pszWalk, szInput))
					return llIterate;

				(unsigned long &) llIterate += 1;
				++pszWalk;
			}

			return LL_NONE;
		}

	protected:
		std::list<BasicLogger *> m_lpLoggers;
		log_level_t m_llFilter;
	};



	class FileLogger : public BasicLogger
	{
	public:
		FileLogger(const char * szFileName, log_level_t llFilter = LL_NONE);
		FileLogger(FILE * pFile, log_level_t llFilter = LL_NONE);
		virtual ~FileLogger();

		virtual void log(log_level_t llLevel, const char * szMessage);
		virtual void setLogLevel(log_level_t llFilter) { m_llFilter = llFilter; }

	protected:
		log_level_t m_llFilter;
		FILE * m_pFile;
		bool m_bClose;
	};
}

#endif // no
