/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * modules.h - header for module stuff ;)
 *
 * 2005-01-05	 wrote this stuff
 */

#include "mwcollectd.h"
#include "log.h"

#ifndef __MWCOLLECTD_MODULE_H
#define __MWCOLLECTD_MODULE_H

#if defined(WIN32)
	#include <windows.h>
#elif defined(LINUX)
	#include <dlfcn.h>
#endif

namespace mwcollectd
{
	class MalwareCollector;

	typedef enum
	{
		MT_GENERIC = 0,
		MT_COLLECTOR,
		MT_SUBMISSION_AGENT,
	} module_type_t;

	class Module
	{
	public:
		virtual void setLogManager(LogManager * pLogManager) = 0;
		virtual void setConfiguration(Configuration * pConfig, const char * szPath) = 0;

		virtual void start() = 0;
		virtual void loop() { }
		virtual void stop() = 0;

		virtual module_type_t moduleType() { return MT_GENERIC; }
	};

	typedef Module * (* module_facility_call_t)(MalwareCollector * pEngine);

	class ModuleRepresentation
	{
	public:
		ModuleRepresentation(const char * szModuleName, mwcollectd::LogManager * pLogger);
		~ModuleRepresentation();

		bool loadLibrary(const char * szFileName);
		bool freeLibrary();

		bool initialiseModule(mwcollectd::MalwareCollector * pCollector, mwcollectd::LogManager * pLogManager, Configuration * pConfig);

		Module * getModule() { return m_pModule; }

	protected:
		char * m_szModuleName;
		mwcollectd::LogManager * m_pLogger;

		#if defined(WIN32)
			HMODULE m_pLibrary;
		#elif defined(LINUX)
			void * m_pLibrary;
		#endif

		module_facility_call_t m_fnCreateInstance;
		Module * m_pModule;
	};
}

#endif // __MWCOLLECTD_MODULE_H
