/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * network.h - networking interface header, basically NetworkInterface + types
 *
 * 2005-01-05	 fast rewrite
 */

#ifndef __MWCOLLECTD_NETWORK_H
#define __MWCOLLECTD_NETWORK_H

#include <vector>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

namespace mwcollectd
{
	typedef void * (* connection_notifier_call_t)(int iHandle, unsigned long ulSourceAddress, unsigned short usSourcePort, unsigned short usLocalPort, void * pUserData);
	typedef void (* connected_notifier_call_t)(int iHandle, unsigned short usDestPort, void * pUserData);
	typedef void (* incoming_notifier_call_t)(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam);
	typedef void (* termination_notifier_call_t)(bool bServer, void * pParam);
	typedef void (* dgram_incoming_notifier_call_t)(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam, unsigned long ulSourceIP, unsigned short usSourcePort);

	typedef enum
	{
		ST_SERVERLISTEN = 0,
		ST_SERVERCLIENT,
		ST_CLIENTCLIENT,
		
		ST_DATAGRAM,

		ST_POLLBITCH,
		ST_DEAD,
	} socket_type_t;

	typedef struct
	{
		int iSysSocket;
		unsigned short usPort;		

		connection_notifier_call_t fnConnectionNotifier;
		incoming_notifier_call_t fnIncomingNotifier;
		dgram_incoming_notifier_call_t fnDatagramNotifier;
		termination_notifier_call_t fnTerminationNotifier;
		socket_type_t stType;
		void * pUserData;
		
		bool bKeepAlive;
		unsigned long ulTimeout;
		unsigned long ulLastAction;
	} socket_t;

	class NetworkInterface
	{
	public:
		NetworkInterface(LogManager * pLogger, Configuration * pConfig);
		virtual ~NetworkInterface();

		void loop();
		void setTimeout(unsigned long ulTimeout) { m_ulTimeout = ulTimeout; }

		virtual int createSocket(unsigned short usPort, connection_notifier_call_t fnConnNotifier, incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData = 0);
		virtual int createUnboundDatagramSocket(dgram_incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData = 0);
		virtual int connectSocket(const char * szHost, unsigned short usPort, connected_notifier_call_t fnConnNotifier, incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData = 0);
		virtual int connectSocket(unsigned long ulAddress, unsigned short usPort, connected_notifier_call_t fnConnNotifier, incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData = 0);
		virtual int addPollSocket(int iSysSocket);

		virtual bool closeSocket(int iHandle, bool bForce = false);

		virtual bool sendData(int iHandle, unsigned char * pucPacket, unsigned int nPacketLen);
		virtual bool sendDatagram(int iHandle, unsigned char * pucPacket, unsigned int nPacketLen, unsigned long ulRemote, unsigned short usPort);

		virtual void setKeepAlive(int iHandle, bool bKeepAlive);
		virtual void setTimeout(int iHandle, unsigned long ulTimeout);
		virtual void timeoutNoise(int iHandle);

		virtual unsigned long lookupHost(const char * szHostname);

		virtual void fastpollHack(bool bEnable) { m_bFastpollHack = bEnable; }

	protected:
		void closeAll();

		void addClient(const socket_t * psckServerSocket, int iNewSysSocket, const struct sockaddr_in * pAddr);

	protected:
		std::list<socket_t> m_lsSockets;

		LogManager * m_pLogManager;
		Configuration * m_pConfig;

		unsigned long m_ulTimeout;

		bool m_bFastpollHack;

		char * m_szPacketBuffer;
	};
}

#endif // __MWCOLLECTD_NETWORK_H
