/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * irc.cpp - irc helper functions for log-irc module
 *
 * 2005-01-06	 fast rewrite
 */

#include "log-irc.h"

void IrcConnectedNotifier(int iHandle, unsigned short usDestPort, void * pUserData)
{
	((irc_state_t *) pUserData)->pLogger->log(mwcollectd::LL_DEBUG, "Connected to IRC Network...");

	{
		char * szUserCommand, * szNickCommand;

		asprintf(&szUserCommand, "USER %s 0 0 :mwcollect report-irc module\r\n", IRC_USER);
		asprintf(&szNickCommand, "NICK %s\r\n", ((irc_state_t *) pUserData)->szNick);

		((irc_state_t *) pUserData)->pInterface->sendData(iHandle, (unsigned char *) szUserCommand, strlen(szUserCommand));
		((irc_state_t *) pUserData)->pInterface->sendData(iHandle, (unsigned char *) szNickCommand, strlen(szNickCommand));

		free(szNickCommand);
		free(szUserCommand);
	}
}

inline char * * SplitWords(const char * szLine, int * pWords)
{	
	const char * j;
	char * * ppWords;
	* pWords = 1;

	for(j = szLine; * j; ++j)
		if(* j == ' ')
			* pWords += 1;

	ppWords = (char * *) malloc(sizeof(char *) * (* pWords));

	{
		char * w;
		int iLen;
		j = szLine;

		for(int i = 0; i < * pWords; ++i)
		{
			iLen = 0;

			for(j = szLine;* j && * j != ' '; ++j)
				++iLen;

			w = ppWords[i] = (char *) malloc(iLen + 1);

			while(* szLine && * szLine != ' ')
				* (w++) = * (szLine++);

			* w = 0;
			++szLine;
		}
	}

	return ppWords;
}

inline void FreeWords(char * * ppWords, int iWords)
{
	for(int i = 0; i < iWords; ++i)
		free(ppWords[i]);

	free(ppWords);
}

inline void SplitUser(const char * szFull, char * * pszNick, char * * pszUser, char * * pszHost)
{
	int iLengths[3];
	const char * j;
	char * k;
	int i;

	iLengths[0] = 1;
	iLengths[1] = 1;
	iLengths[2] = 1;

	for(i = 0, j = szFull; * j; ++j)
		if(* j == '!')
			i = 1;
		else if(* j == '@')
			i = 2;
		else
			++iLengths[i];
		
	* pszNick = (char *) malloc(iLengths[0]);
	* pszUser = (char *) malloc(iLengths[1]);
	* pszHost = (char *) malloc(iLengths[2]);

	for(j = szFull, k = * pszNick; * j; ++j)
	{
		if(* j == '!')
		{
			* k = 0;
			k = * pszUser;
		}
		else if(* j == '@')
		{
			* k = 0;
			k = * pszHost;
		}
		else
			* (k++) = *j;
	}

	* k = 0;
}

bool IrcParseLine(const char * szLine, irc_state_t * pState, int iSocket)
{
	int iWords;
	char * * ppWords;
	char * szNick, * szUser, * szHost;
	char * szMessage = strchr(szLine + 1, ':') + 1;
	bool bMoreFollowing = true;

	ppWords = SplitWords(szLine, &iWords);
	SplitUser((ppWords[0][0] == ':' ? ppWords[0] + 1 : ppWords[0]), &szNick, &szUser, &szHost);

	if(iWords < 2)
		goto end;

	{
		if(!strcmp(ppWords[0], "PING"))
		{
			char * szResponse;
			unsigned int nLen;

			nLen = asprintf(&szResponse, "PONG %s\r\n", ppWords[1]);
			pState->pInterface->sendData(iSocket, (unsigned char *) szResponse, nLen);
			free(szResponse);
		}
		else if(!strcmp(ppWords[1], "376") || !strcmp(ppWords[1], "422"))
		{
			if(!pState->fOnChannel)
			{
				char * szJoinCommand;
				int iLen;
				
				if(pState->szKey && strlen(pState->szKey))
					iLen = asprintf(&szJoinCommand, "JOIN %s %s\r\n", pState->szChannel, pState->szKey);
				else
					iLen = asprintf(&szJoinCommand, "JOIN %s\r\n", pState->szChannel);

				pState->pInterface->sendData(iSocket, (unsigned char *) szJoinCommand, iLen);
				free(szJoinCommand);

				pState->fOnChannel = true;

				pState->pModule->startForwarding();
			}			
		}
		else if(!strcmp(ppWords[1], "302"))
		{
			char * szBuffer;

			if(strrchr(ppWords[iWords - 2], '@'))
			{
				int iLen = asprintf(&szBuffer, "PRIVMSG %s :%s\r\n", pState->szChannel, strrchr(ppWords[iWords - 2], '@') + 1);

				pState->pInterface->sendData(iSocket, (unsigned char *) szBuffer, iLen);
				free(szBuffer);
			}
			else
				pState->pLogger->log(mwcollectd::LL_CRITICAL, "Malicious IRC Server, responded badly to USERHOST!");
		}
		else if(!strcmp(ppWords[1], "PRIVMSG") && !strcmp(pState->szChannel, ppWords[2]))
		{
			bool fAdmin = false;

			if((!strcmp(pState->szAdminNick, "*") || !strcmp(pState->szAdminNick, szNick)) && (!strcmp(pState->szAdminUser, "*") || !strcmp(pState->szAdminUser, szUser)) && (!strcmp(pState->szAdminHost, "*") || !strcmp(pState->szAdminHost, szHost)))
				fAdmin = true;

			if(fAdmin)
			{
				if(!strcmp(szMessage, ".quit"))
				{
					pState->pLogger->log(mwcollectd::LL_INFO, "%s ordered IRC Bot to quit.", ppWords[0] + 1);
					pState->pModule->stop();

					bMoreFollowing = false;
				}
				else if(!strncmp(szMessage, ".level ", 7))
				{
					if(mwcollectd::LogManager::mapLogLevel(szMessage + 7) == mwcollectd::LL_NONE)
					{
						char * szErrorMessage;
						int iLen = asprintf(&szErrorMessage, "PRIVMSG %s :Unknown log level \"%s\".\r\n", ppWords[2], szMessage + 7);

						pState->pInterface->sendData(iSocket, (unsigned char *) szErrorMessage, iLen);
						free(szErrorMessage);
					}
					else
					{
						char * szConfirmMessage;
						int iLen = asprintf(&szConfirmMessage, "PRIVMSG %s :Switched to log level \"%s\".\r\n", ppWords[2], szMessage + 7);

						pState->pInterface->sendData(iSocket, (unsigned char *) szConfirmMessage, iLen);
						free(szConfirmMessage);

						pState->pModule->setLogLevel(mwcollectd::LogManager::mapLogLevel(szMessage + 7));
					}
				}
				else if(!strcmp(szMessage, ".ip"))
				{
					char * szBuffer;
					int iLen = asprintf(&szBuffer, "USERHOST %s\r\n", pState->szNick);

					pState->pInterface->sendData(iSocket, (unsigned char *) szBuffer, iLen);
					free(szBuffer);
				}
			}

			if(!strcmp(szMessage, ".who's-your-daddy"))
			{
				char * szConfirmMessage;
				int iLen = asprintf(&szConfirmMessage, "PRIVMSG %s :%s!%s@%s may grab my ass, but oxff is my true daddy!\r\n", ppWords[2], pState->szAdminNick, pState->szAdminUser, pState->szAdminHost);

				pState->pInterface->sendData(iSocket, (unsigned char *) szConfirmMessage, iLen);
				free(szConfirmMessage);
			}
		}
	}

end:
	FreeWords(ppWords, iWords);

	free(szNick);
	free(szUser);
	free(szHost);

	return bMoreFollowing;
}

void IrcIncomingNotifier(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam)
{
	irc_state_t * pState = (irc_state_t *) pParam;

	for(unsigned int i = 0; i < nPacketLength; ++i)
	{		
		if(pState->usBufferPos > 1 && pState->szBuffer[pState->usBufferPos - 2] == '\r' && pState->szBuffer[pState->usBufferPos - 1] == '\n')
		{
			pState->szBuffer[pState->usBufferPos - 2] = 0;

			if(IrcParseLine(pState->szBuffer, pState, iHandle))
				pState->usBufferPos = 0;
		}

		if(pState->usBufferPos > 514)
		{
			pState->pLogger->log(mwcollectd::LL_CRITICAL, "IRC Line Overflow in " __FILE__ ":%u!", __LINE__);
			return;
		}

		pState->szBuffer[pState->usBufferPos++] = pucPacketData[i];
	}

	if(pState->usBufferPos > 1 && pState->szBuffer[pState->usBufferPos - 2] == '\r' && pState->szBuffer[pState->usBufferPos - 1] == '\n')
	{
		pState->szBuffer[pState->usBufferPos - 2] = 0;

		if(IrcParseLine(pState->szBuffer, pState, iHandle))
			pState->usBufferPos = 0;
	}
}

void IrcTerminationNotifier(bool bServer, void * pParam)
{
	irc_state_t * pState = (irc_state_t *) pParam;

	pState->pModule->connectionLost();
	pState->pLogger->log(mwcollectd::LL_INFO, "Lost connection to IRC Server.");

	free(pState->szChannel);
	free(pState->szKey);
	free(pState->szPassword);
	free(pState->szNick);

	free(pState->szAdminNick);
	free(pState->szAdminUser);
	free(pState->szAdminHost);

	free(pParam);
}
