/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * module.cpp - Module class for log-irc
 *
 * 2005-01-06	 fast rewrite
 */

#include "log-irc.h"

extern "C"
{
  mwcollectd::Module * __attribute__ ((dllexport)) CreateInstance(mwcollectd::MalwareCollector * pCollector)
  {
    return (new ReportIrcModule(pCollector->getNetworkInterface()));
  }
}


ReportIrcModule::ReportIrcModule(mwcollectd::NetworkInterface * pNetIface)
{
  m_pLogger = 0;
  m_pConfig = 0;
  m_szConfigPath = 0;
  m_pNetworkInterface = pNetIface;
  m_iSocket = -1;
  m_szChannel = 0;

  m_ulNextTry = 0;
}

ReportIrcModule::~ReportIrcModule()
{
	if(m_szConfigPath)
		free(m_szConfigPath);
}

void ReportIrcModule::setLogManager(mwcollectd::LogManager * pLogger)
{
  m_pLogger = pLogger;

  m_llLevel = LL_NONE;
}

void ReportIrcModule::setConfiguration(Configuration * pConfig, const char * szPath)
{
  m_pConfig = pConfig;
  m_szConfigPath = strdup(szPath);

  {
	  char * szPath2;
	  
	  asprintf(&szPath2, "%s:server:channel", szPath);
	  m_szChannel = m_pConfig->getString(szPath2, "#mwcollect");
	  free(szPath2);
  }
}

void ReportIrcModule::start()
{
  irc_state_t * pIrcState = (irc_state_t *) malloc(sizeof(irc_state_t));

  m_bRetry = true;

  pIrcState->pModule = this;  
  pIrcState->pLogger = m_pLogger;
  pIrcState->pConfig = m_pConfig;
  pIrcState->pInterface = m_pNetworkInterface;

  pIrcState->szBuffer[0] = 0;
  pIrcState->usBufferPos = 0;
  pIrcState->fOnChannel = false;
  
  {
	  const char * szHost, * szChannel, * szKey, * szPassword, * szNick;
	  const char * szAdminNick, * szAdminUser, * szAdminHost;
	  char * szPath;
	  long lPort;

	  {
		  asprintf(&szPath, "%s:server:host", m_szConfigPath);
		  szHost = m_pConfig->getString(szPath, "127.0.0.1");
		  free(szPath);

		  asprintf(&szPath, "%s:server:channel", m_szConfigPath);
		  szChannel = m_pConfig->getString(szPath, "#mwcollect");
		  free(szPath);

		  asprintf(&szPath, "%s:server:key", m_szConfigPath);
		  szKey = m_pConfig->getString(szPath, "\0");
		  free(szPath);

		  asprintf(&szPath, "%s:server:password", m_szConfigPath);
		  szPassword = m_pConfig->getString(szPath, "\0");
		  free(szPath);

		  asprintf(&szPath, "%s:server:port", m_szConfigPath);
		  lPort = m_pConfig->getLong(szPath, 6667);
		  free(szPath);

		  asprintf(&szPath, "%s:nick", m_szConfigPath);
		  szNick = m_pConfig->getString(szPath, "mwcollect");
		  free(szPath);


		  asprintf(&szPath, "%s:admin:nick", m_szConfigPath);
		  szAdminNick = m_pConfig->getString(szPath, "*");
		  free(szPath);

		  asprintf(&szPath, "%s:admin:user", m_szConfigPath);
		  szAdminUser = m_pConfig->getString(szPath, "*");
		  free(szPath);

		  asprintf(&szPath, "%s:admin:host", m_szConfigPath);
		  szAdminHost = m_pConfig->getString(szPath, "*");
		  free(szPath);
	  }

	  pIrcState->szChannel = strdup(szChannel);
	  pIrcState->szKey = strdup(szKey);
	  pIrcState->szPassword = strdup(szPassword);
	  pIrcState->szNick = strdup(szNick);
	  
	  pIrcState->szAdminUser = strdup(szAdminUser);
	  pIrcState->szAdminNick = strdup(szAdminNick);
	  pIrcState->szAdminHost = strdup(szAdminHost);

	  if((m_iSocket = m_pNetworkInterface->connectSocket(szHost, lPort, IrcConnectedNotifier, IrcIncomingNotifier, IrcTerminationNotifier, pIrcState)) < 0)
	  {
		m_pLogger->log(LL_CRITICAL, "Could not connect to IRC Server %s:%u!", szHost, lPort);
		m_ulNextTry = time(0) + (60 * 2);
	  }
	  else
	  {
		  m_pNetworkInterface->setTimeout(m_iSocket, 600); // ten minutes should be fairly enough as a irc server should send PINGs..

		  m_ulNextTry = 0;
	  }
  }
}

void ReportIrcModule::loop()
{
	if(m_bRetry && m_ulNextTry && m_ulNextTry >= (unsigned long) time(0))
		start();
}

void ReportIrcModule::stop()
{
  int iLen;
  char * szQuitMessage;

  iLen = asprintf(&szQuitMessage, "QUIT :mwcollect2 exiting\r\n");
  m_pNetworkInterface->sendData(m_iSocket, (unsigned char *) szQuitMessage, iLen);
  free(szQuitMessage);

  m_bRetry = false;

  m_pNetworkInterface->closeSocket(m_iSocket);
  m_iSocket = -1;

  m_pLogger->removeLogger(this);
}

void ReportIrcModule::log(mwcollectd::log_level_t llLevel, const char * szMessage)
{
	char * pszColourTable[] = { "", "\x03\x32", "\x03\x33", "\x03\x38", "\x02\x03\x34" };

  if(llLevel < m_llLevel || m_iSocket < 0 || (unsigned int) llLevel >= sizeof(pszColourTable) / sizeof(char *))
    return;

  char * szBuffer;
  int iLen = asprintf(&szBuffer, "PRIVMSG %s :%s%s\r\n", m_szChannel, pszColourTable[llLevel], szMessage);

  m_pNetworkInterface->sendData(m_iSocket, (unsigned char *) szBuffer, iLen);
  m_pNetworkInterface->timeoutNoise(m_iSocket);
  free(szBuffer);
}

void ReportIrcModule::startForwarding()
{
  m_pLogger->addLogger(this);
}

void ReportIrcModule::connectionLost()
{
	m_pLogger->removeLogger(this);

	m_ulNextTry = time(0) + (60 * 2);
}
