/* mwcollect2 (c) 2004-2005 Georg Wicherski
 * this file is not licensed under the license in the file LICENSE but is public domain
 * THIS IS THE _ONLY_ PUBLIC DOMAIN FILE IN THIS PROJECT
 *
 * config.cpp - config file parsing (Configuration implementation)
 *
 * 2005-01-05	coded this bloody parsing biatch
 * 2005-01-06	'minimzed c' (only five syscalls per file left)
 */

// uncomment the following line if you want to use this in your own project
// #define STANDALONE_MWCOLLECT_CONFIG_FILE
// if you want to seperate the definition into a header, cut the stuff between the #ifdef into a header

#include <list>
#include <string>
#include <sys/stat.h>

#ifdef STANDALONE_MWCOLLECT_CONFIG_FILE

	typedef struct configuration_node_s
	{
		std::string sName;

		struct // was union but not allowed because of constructor.. :/
		{
			std::string sValue;
			std::list<configuration_node_s *> lpChildren;
		} uData;

		configuration_node_s * pFather; // needed for parsing
		bool bIsLeaf; // has sValue or lpChildren?; an empty block is considered as not beeing leaf!
	} configuration_node_t;

	class Configuration
	{
	public:
		Configuration();
		~Configuration();

		bool parse(const char * szFile);
		void free();

		const char * getString(const char * szPath, const char * szDefault);
		long getLong(const char * szPath, long lDefault);
		std::list<const char *> * getChildren(const char * szPath);
		bool leafExists(const char * szPath);
		bool blockExists(const char * szPath);

		void statistics(unsigned int& pBlocks, unsigned int& pValues, unsigned int& pCharacters);

	private:
		void freenode(configuration_node_t * pFree); // yeah I love you irc guys ;P

		configuration_node_t * pathToNode(const char * szPath, configuration_node_t * pStart = 0);

	protected:
		configuration_node_t * m_pRoot;

		unsigned int m_nBlocks, m_nValues, m_nCharacters; // statisitics are always cool
	};

#else

	#include <mwcollectd.h>

	namespace mwcollectd
	{

#endif

// add your complain callback / whatever here or just define it empty
#define COMPLAIN(a) printf("[parse error] line %u, column %u, parsing '%c', [" __FILE__ ":%u]: %s\n", y, x, c, __LINE__, a)


Configuration::Configuration()
{
	m_pRoot = 0;
}

Configuration::~Configuration()
{
	free();
}


void Configuration::free()
{
	freenode(m_pRoot);
}

void Configuration::freenode(configuration_node_t * pFree)
{
	if(pFree->bIsLeaf)
	{
		for(std::list<configuration_node_t *>::iterator i = pFree->uData.lpChildren.begin(); i != pFree->uData.lpChildren.end(); ++i)
			freenode(* i);
	}

	delete pFree;
}

typedef enum
{ // different states for our state parser ::parse
	CPS_IN_VOID,
	CPS_IN_CHILD,

	CPS_IN_NAME,
	CPS_IN_VALUE,
	CPS_IN_VALUE_ESCAPE,

	CPS_EXPECTING_VEND,
	CPS_EXPECTING_VALUE,
	
	CPS_IN_COMMENT,
} cfg_parse_state_t;

bool Configuration::parse(const char * szFile)
{ // bloody biaaatch ;)
	FILE * pFile;
	cfg_parse_state_t cpsState, cpsRestore;
	unsigned int nDepth, nRead = 0;
	std::string sName, sValue;
	int x = 0, y = 1;
	configuration_node_t * pCurrent;
	char * szBuffer;

	{
		struct stat fs;

		if(m_pRoot)
			return false;

		if(!(pFile = fopen(szFile, "rb")))
			return false;

		stat(szFile, &fs);

		if(!(szBuffer = (char *) malloc(fs.st_size)))
			return false;

		if(fread(szBuffer, 1, fs.st_size, pFile) < 1)
			return false;

		fclose(pFile);

		m_nCharacters = (unsigned int) fs.st_size;
	}

	{
		m_pRoot = new configuration_node_t;

		m_pRoot->uData.lpChildren.clear();
		m_pRoot->bIsLeaf = false;
		m_pRoot->pFather = 0;
		m_pRoot->sName = std::string("root");

		cpsState = CPS_IN_VOID;

		m_nBlocks = 0;
		m_nValues = 0;
	}

	{ // actual parsing
		char c;

		for(nRead = 0; nRead < m_nCharacters; ++nRead)
		{
			c = szBuffer[nRead];

			if(c == '\n')
			{
				++y; x = 0;
			}
			else if(c != '\r')
				++x;			

			// these are more or less general rules for comments and whitespace characters, they apply to most states

			if(cpsState == CPS_IN_COMMENT)
			{
				if(c == '\n')
					cpsState = cpsRestore;
				else
					continue;
			}

			if(c == ';' && cpsState != CPS_IN_VALUE && cpsState != CPS_EXPECTING_VEND) // outside of values, slightly ignored if not required
				continue;

			if(cpsState != CPS_IN_VALUE && isspace(c))
				continue; // if we're not currently parsing some value, discard any whitespace characters

			if(cpsState != CPS_IN_VALUE && c == '#')
			{
				cpsRestore = cpsState;
				cpsState = CPS_IN_COMMENT;

				continue;
			}

			
			// these are state specific rules

			switch(cpsState)
			{
			case CPS_IN_VOID:
				if(c != '{')
				{
					COMPLAIN("Parsing in void, expecting open root bracket or comment.");
					return false;
				}
				else
				{
					sName.erase();
					nDepth = 1;

					pCurrent = m_pRoot;
					cpsState = CPS_IN_CHILD;
				}

				break;

			case CPS_IN_CHILD:
				if(c == '}')
				{
					--nDepth;

					if(!nDepth)
						cpsState = CPS_IN_VOID;
					else
						pCurrent = pCurrent->pFather;

					continue;
				}
				else
					cpsState = CPS_IN_NAME;

			case CPS_IN_NAME:
				if(c == '=')
					cpsState = CPS_EXPECTING_VALUE;
				else if(c == '{')
				{ // create a new child
					configuration_node_t * pNew = new configuration_node_t;

					if(pCurrent->bIsLeaf)
					{
						COMPLAIN("Want to create a block child for a leaf!");
						return false;
					}

					pNew->pFather = pCurrent;
					pNew->sName = sName;
					pNew->uData.lpChildren.clear();
					pNew->bIsLeaf = false;

					pCurrent->uData.lpChildren.push_back(pNew);
					pCurrent = pNew;

					sName.erase();

					++nDepth;
					++m_nBlocks;
					cpsState = CPS_IN_CHILD;
				}
				else
					sName.push_back(c);

				break;

			case CPS_EXPECTING_VALUE:
				if(c == '\"')
				{
					sValue.erase();
					cpsState = CPS_IN_VALUE;
				}
				else
				{
					COMPLAIN("Expecting Value, however found did not find \".");
					return false;
				}

				break;

			case CPS_IN_VALUE:
				if(c == '\"')
				{
					configuration_node_t * pNew = new configuration_node_t;

					if(pCurrent->bIsLeaf)
					{
						COMPLAIN("Want to create a leaf child for a leaf!");
						return false;
					}

					pNew->sName = sName;
					pNew->uData.sValue = sValue;
					pNew->bIsLeaf = true;
					pNew->pFather = pCurrent;

					sValue.erase();
					sName.erase();

					pCurrent->uData.lpChildren.push_back(pNew);
					// no pCurrent = pNew needed since we have a leaf and so we are continuing in the old one

					cpsState = CPS_EXPECTING_VEND;
					++m_nValues;
				}
				else if(c == '\\')
					cpsState = CPS_IN_VALUE_ESCAPE;
				else
					sValue.push_back(c);

				break;

			case CPS_IN_VALUE_ESCAPE: // yeah we are nearly like a C++ parser... nearly... at least a bit...
				if(c == '\"')
					sValue.push_back('\"');
				else if(c == 't')
					sValue.push_back('\t');
				else if(c == 'n')
					sValue.push_back('\n');
				else if(c == 'r')
					sValue.push_back('\r');
				else if(c == '\\')
					sValue.push_back('\\');
				else
				{
					COMPLAIN("Unknown escape character.");
					return false;
				}

				cpsState = CPS_IN_VALUE;
				break;

			case CPS_EXPECTING_VEND:
				if(c == ';')
					cpsState = CPS_IN_CHILD;
				else
				{
					COMPLAIN("Unterminated value.. missing ';'?");
					return false;
				}

				break;

			default:
				COMPLAIN("Unknown state in parser!");
				return false;
			}
		}
	}

	::free(szBuffer);
	return true;
}


configuration_node_t * Configuration::pathToNode(const char * szPath, configuration_node_t * pStart)
{ // ok we get something like :foo:bar:baz and turn this into the pointer to the struct representing the baz node, using recursion
	std::string sName;

	if(!pStart)
	{
		if(* szPath == ':')
			++szPath;

		pStart = m_pRoot;
	}

	if(pStart->bIsLeaf)
		return 0;

	while(* szPath && * szPath != ':')
		sName.push_back(* (szPath++));

	for(std::list<configuration_node_t *>::iterator i = pStart->uData.lpChildren.begin(); i != pStart->uData.lpChildren.end(); ++i)
	{
		if((* i)->sName != sName)
			continue;
		else if(* szPath)
			return pathToNode(++szPath, (* i));
		else
			return (* i);
	}

	return 0; // no such foobarian
}

bool Configuration::blockExists(const char * szPath)
{ // test if the given path leads to a valid block
	configuration_node_t * pNode = pathToNode(szPath);

	if(!pNode || pNode->bIsLeaf)
		return false;

	return true;
}

bool Configuration::leafExists(const char * szPath)
{ // test if the given path leads to a valid leaf
	configuration_node_t * pNode = pathToNode(szPath);

	if(!pNode || !pNode->bIsLeaf)
		return false;

	return true;
}

std::list<const char *> * Configuration::getChildren(const char * szPath)
{ // enumerate all the children of a node represented by a given task, this isn't that complicated in general,
	// however needs some work to get it into some ['programmer friendly' | 'manipulation safe'] <const char *>list...
	// RESULT IS NOT THREAD SAFE because next call will modify the result list, next parsing will modify the pointers in the list

	static std::list<const char *> lszReturn;
	configuration_node_t * pNode = pathToNode(szPath);

	if(!pNode || pNode->bIsLeaf)
		return 0;

	lszReturn = std::list<const char *>();

	for(std::list<configuration_node_t *>::iterator i = (pNode->uData.lpChildren).begin(); i != (pNode->uData.lpChildren).end(); ++i)
		lszReturn.push_back((* i)->sName.c_str());

	return &lszReturn;
}

const char * Configuration::getString(const char * szPath, const char * szDefault)
{ // fetch a string value by path
	// MORE OR LESS THREAD SAFE, next parsing will modify the result pointer

	configuration_node_t * pNode = pathToNode(szPath);

	if(!pNode || !pNode->bIsLeaf)
		return szDefault;

	return pNode->uData.sValue.c_str();
}

long Configuration::getLong(const char * szPath, long lDefault)
{ // fetch a long by path
	// THREADSAFE since returned as copy
	// if leaf contains non-number string literal instead of some number, this will return 0 instead of lDefault so take care

	configuration_node_t * pNode = pathToNode(szPath);

	if(!pNode || !pNode->bIsLeaf)
		return lDefault;

	return strtol(pNode->uData.sValue.c_str(), 0, 0);
}

void Configuration::statistics(unsigned int& pBlocks, unsigned int& pValues, unsigned int& pCharacters)
{
	pBlocks = m_nBlocks;
	pValues = m_nValues;
	pCharacters = m_nCharacters;
}

#ifndef STANDALONE_MWCOLLECT_CONFIG_FILE

	} // close the namespace

#endif
