/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * log.cpp - implementation of LogManager and FileLogger
 *
 * 2005-01-05	wrote LogManager and FileLogger
 *
 */

#include <mwcollectd.h>

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

namespace mwcollectd
{
	LogManager::LogManager(log_level_t llDefaultFilter)
	{
		m_llFilter = llDefaultFilter;
		m_lpLoggers.clear();
	}

	LogManager::~LogManager()
	{
	}

	void LogManager::addLogger(BasicLogger * pLogger)
	{
		if(m_llFilter)
			pLogger->setLogLevel(m_llFilter);

		m_lpLoggers.push_back(pLogger);
	}

	void LogManager::removeLogger(BasicLogger * pLogger)
	{
		m_lpLoggers.remove(pLogger);
	}

	std::list<BasicLogger *> * LogManager::loggerList()
	{
		return &m_lpLoggers;
	}

	void LogManager::log(log_level_t llLevel, const char * szFormat, ...)
	{ // parses the format foo and sends it to all listed client loggers
		char * szMessage;
		va_list vaArguments;

		va_start(vaArguments, szFormat);
		vasprintf(&szMessage, szFormat, vaArguments);
		va_end(vaArguments);

		// iterate
		for(std::list<BasicLogger *>::iterator i = m_lpLoggers.begin(); i != m_lpLoggers.end(); ++i)
			(* i)->log(llLevel, szMessage);

		free(szMessage);
	}



	FileLogger::FileLogger(const char * szFileName, log_level_t llFilter)
	{
		m_llFilter = llFilter;
		m_pFile = fopen(szFileName, "at");
		m_bClose = true;

		chmod(szFileName, S_IRUSR | S_IWUSR | S_IRGRP); // savety first.
	}

	FileLogger::FileLogger(FILE * pFile, log_level_t llFilter)
	{ // usefull for passing stdin / stderr for example
		m_llFilter = llFilter;
		m_pFile = pFile;
		m_bClose = false;
	}

	FileLogger::~FileLogger()
	{
		if(m_bClose && m_pFile)
			fclose(m_pFile);

		m_pFile = 0;
	}


	void FileLogger::log(log_level_t llLevel, const char * szMessage)
	{ // write the line to the file descriptor
		struct tm t;
		time_t tStamp;

		if(llLevel < m_llFilter)
			return;

		time(&tStamp);
		localtime_r(&tStamp, &t);

		if(m_pFile)
			fprintf(m_pFile, "[%1u %02d-%02d-%04d %02d:%02d:%02d] %s\n", llLevel, t.tm_mday, t.tm_mon + 1, t.tm_year + 1900, t.tm_hour, t.tm_min, t.tm_sec, szMessage);

		fflush(m_pFile);
	}
}
