/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * network.cpp - networking interface including callback calls
 *
 * 2005-01-04	wrote da shit =)
 * 2005-01-18	added lookupHost
 * 2005-03-10	complete rewrite
 */

#include <mwcollectd.h>
#include <time.h>
#include <list>
#include <unistd.h>

#define PACKET_BUFFER_SIZE 4096

namespace mwcollectd
{
	NetworkInterface::NetworkInterface(LogManager * pLogger, Configuration * pConfig)
	{ // simple constructor which initalizes member variables
		m_pLogManager = pLogger;
		m_pConfig = pConfig;

		m_bFastpollHack = false;
		m_ulTimeout = m_pConfig->getLong(":data-timeout", 300);

		m_szPacketBuffer = (char *) malloc(PACKET_BUFFER_SIZE);
	}

	NetworkInterface::~NetworkInterface()
	{
		free(m_szPacketBuffer);
	}


	int NetworkInterface::createSocket(unsigned short usPort, connection_notifier_call_t fnConnNotifier, incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData)
	{ // creates a listening socket bound to a specified port and adds to poll list
		struct sockaddr_in addrLocal;
		socket_t sckSocket;

		addrLocal.sin_family = AF_INET;
		addrLocal.sin_port = htons(usPort);
		addrLocal.sin_addr.s_addr = INADDR_ANY;

		if((sckSocket.iSysSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
			return -1;

		if(bind(sckSocket.iSysSocket, (struct sockaddr *) &addrLocal, sizeof(addrLocal)) < 0)
		{
			m_pLogManager->log(LL_CRITICAL, "Binding to :%u failed.", usPort);

			return -1;
		}

		if(listen(sckSocket.iSysSocket, m_pConfig->getLong(":backlog", 16)) < 0)
			return -1;

		sckSocket.usPort = usPort;

		sckSocket.fnConnectionNotifier = fnConnNotifier;
		sckSocket.fnIncomingNotifier = fnInNotifier;
		sckSocket.fnTerminationNotifier = fnTermNotifier;
		sckSocket.fnDatagramNotifier = 0;
		
		sckSocket.stType = ST_SERVERLISTEN;
		sckSocket.pUserData = pUserData;

		sckSocket.bKeepAlive = true;
		sckSocket.ulTimeout = m_ulTimeout;
		sckSocket.ulLastAction = time(0);

		m_lsSockets.push_back(sckSocket);
		return sckSocket.iSysSocket;
	}

	int NetworkInterface::connectSocket(const char * szHost, unsigned short usPort, connected_notifier_call_t fnConnNotifier, incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData)
	{ // wrapper for lazy ppl ;)
		return connectSocket(lookupHost(szHost), usPort, fnConnNotifier, fnInNotifier, fnTermNotifier, pUserData);
	}

	int NetworkInterface::connectSocket(unsigned long ulAddress, unsigned short usPort, connected_notifier_call_t fnConnNotifier, incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData)
	{ // connects (blocking!) to a given server and adds socket to poll list
		struct sockaddr_in addrRemote;
		socket_t sckSocket;

		addrRemote.sin_family = AF_INET;
		addrRemote.sin_port = htons(usPort);
		addrRemote.sin_addr.s_addr = ulAddress;

		if((sckSocket.iSysSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
			return -1;

		if(connect(sckSocket.iSysSocket, (struct sockaddr *) &addrRemote, sizeof(addrRemote)) < 0)
		{
			m_pLogManager->log(LL_INFO, "Connecting to %s:%u failed!", inet_ntoa(* (in_addr *) &ulAddress), usPort);

			return -1;
		}

		sckSocket.stType = ST_CLIENTCLIENT;
		sckSocket.usPort = usPort;

		sckSocket.fnIncomingNotifier = fnInNotifier;
		sckSocket.fnTerminationNotifier = fnTermNotifier;
		sckSocket.fnDatagramNotifier = 0;
		sckSocket.fnConnectionNotifier = 0;
		
		sckSocket.pUserData = pUserData;

		sckSocket.bKeepAlive = true;
		sckSocket.ulTimeout = m_ulTimeout;
		sckSocket.ulLastAction = time(0);

        m_pLogManager->log(LL_SPAM, "Connected Socket.");

        fnConnNotifier(sckSocket.iSysSocket, usPort, pUserData);
		m_lsSockets.push_back(sckSocket);
		return sckSocket.iSysSocket;
	}

	int NetworkInterface::createUnboundDatagramSocket(dgram_incoming_notifier_call_t fnInNotifier, termination_notifier_call_t fnTermNotifier, void * pUserData)
	{ // creates an unbound udp socket and adds it to the poll list
		socket_t sckSocket;

		if((sckSocket.iSysSocket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
			return -1;

		sckSocket.stType = ST_DATAGRAM;
		sckSocket.usPort = 0;

		sckSocket.fnConnectionNotifier = 0;
		sckSocket.fnIncomingNotifier = 0;
		sckSocket.fnDatagramNotifier = fnInNotifier;
		sckSocket.fnTerminationNotifier = fnTermNotifier;
		
		sckSocket.pUserData = pUserData;

		sckSocket.bKeepAlive = false;
		sckSocket.ulTimeout = m_ulTimeout;
		sckSocket.ulLastAction = time(0);

		m_lsSockets.push_back(sckSocket);
		return sckSocket.iSysSocket;
	}

	void NetworkInterface::addClient(const socket_t * psckServerSocket, int iNewSysSocket, const struct sockaddr_in * pAddr)
	{ // protected helper function that creates a poll list entry for a connected client
		socket_t sckSocket;

		sckSocket.iSysSocket = iNewSysSocket;
		sckSocket.stType = ST_SERVERCLIENT;
		sckSocket.usPort = psckServerSocket->usPort;

		sckSocket.fnConnectionNotifier = 0;
		sckSocket.fnIncomingNotifier = psckServerSocket->fnIncomingNotifier;
		sckSocket.fnDatagramNotifier = 0;
		sckSocket.fnTerminationNotifier = psckServerSocket->fnTerminationNotifier;

		sckSocket.pUserData = psckServerSocket->fnConnectionNotifier(iNewSysSocket, pAddr->sin_addr.s_addr, ntohs(pAddr->sin_port), psckServerSocket->usPort, psckServerSocket->pUserData);

		sckSocket.bKeepAlive = false;
		sckSocket.ulTimeout = m_ulTimeout;
		sckSocket.ulLastAction = time(0);

		m_lsSockets.push_back(sckSocket);
	}


	bool NetworkInterface::closeSocket(int iHandle, bool bForce)
	{ // close a socket and remove it from the poll list (~ mark it as dead and wait for loop)

		for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i)
		{
			if(i->iSysSocket == iHandle)
			{
				if(i->stType != ST_POLLBITCH && !bForce)
				{
					shutdown(i->iSysSocket, SHUT_RDWR);
					close(i->iSysSocket);
				}

				i->stType = ST_DEAD;
				return true;
			}
		}

		return false;
	}

	void NetworkInterface::closeAll()
	{ // protected helper function to close all sockets on poll list (-> destructor)
		for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i)
		{
			if(i->stType != ST_POLLBITCH)
			{
				shutdown(i->iSysSocket, SHUT_RDWR);
				close(i->iSysSocket);
			}
		}

		m_lsSockets.clear();
	}
	

	int NetworkInterface::addPollSocket(int iSysSocket)
	{ // add a socket only for breaking the poll call (e.g. used by postgresql)
		socket_t sckSocket = { 0 };

		sckSocket.iSysSocket = iSysSocket;
		sckSocket.stType = ST_POLLBITCH;

		m_lsSockets.push_back(sckSocket);
		return iSysSocket;
	}


	bool NetworkInterface::sendData(int iHandle, unsigned char * pucPacket, unsigned int nPacketLen)
	{
		return (send(iHandle, (char *) pucPacket, nPacketLen, 0) == (int) nPacketLen);
	}

	bool NetworkInterface::sendDatagram(int iHandle, unsigned char * pucPacket, unsigned int nPacketLen, unsigned long ulRemote, unsigned short usPort)
	{
		struct sockaddr_in addrRemote;

		addrRemote.sin_addr.s_addr = ulRemote;
		addrRemote.sin_port = htons(usPort);
		addrRemote.sin_family = AF_INET;

		return (sendto(iHandle, (char *) pucPacket, nPacketLen, 0, (struct sockaddr *) &addrRemote, sizeof(addrRemote)) == (int) nPacketLen);
	}


	void NetworkInterface::loop()
	{ // called every iteration; blocks main collector if no net data is incoming
		unsigned long ulTime = time(0);

		{ // check for socket timeouts
			for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i)
			{
				if(i->stType != ST_POLLBITCH && (!i->bKeepAlive && ulTime - i->ulLastAction > i->ulTimeout))
				{
					shutdown(i->iSysSocket, SHUT_RDWR);
					close(i->iSysSocket);

					i->stType = ST_DEAD;
				}
			}
		}

		{ // remove dead sockets
			for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end();)
			{
				if(i->stType == ST_DEAD)
                {
					i->fnTerminationNotifier(i->stType == ST_SERVERLISTEN, i->pUserData);

					m_lsSockets.erase(i);
					i = m_lsSockets.begin();
				}
				else
					++i;
			}
		}

		{ // poll for interesting events and take appropriate actions
			struct pollfd * pDescriptors = (struct pollfd *) malloc(sizeof(struct pollfd) * m_lsSockets.size());
			unsigned int j = 0;

			{ // get status for each socket and wait if there is nothing to do
				for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i, ++j)
				{
					pDescriptors[j].fd = i->iSysSocket;
					pDescriptors[j].events = POLLIN | POLLERR;
					pDescriptors[j].revents = 0;
				}

				if(poll(pDescriptors, j, m_bFastpollHack ? 5 : 2000) < 0)
				{
					free(pDescriptors);

					return;
				}
			}

			{ // take appropriate actions
				std::list<socket_t> lsSaveCopy = m_lsSockets; // a notifier could add sockets, change list ordering and therefore screw j
				j = 0;

				for(std::list<socket_t>::iterator i = lsSaveCopy.begin(); i != lsSaveCopy.end(); ++i, ++j)
				{
					if(i->stType == ST_DEAD)
						continue;

					if(pDescriptors[j].revents & POLLERR)
					{
                        m_pLogManager->log(LL_SPAM, "Poll Error on %i (0x%08X).", i->iSysSocket, i->pUserData);

                        closeSocket(i->iSysSocket, true);
					}
					else if(pDescriptors[j].revents & POLLIN)
					{
						switch(i->stType)
						{
						case ST_POLLBITCH:
							break;

						case ST_SERVERCLIENT:
						case ST_CLIENTCLIENT:
							{
								int iLength = recv(i->iSysSocket, m_szPacketBuffer, PACKET_BUFFER_SIZE, 0);

								if(iLength <= 0)
                                    closeSocket(i->iSysSocket);
                                else
									i->fnIncomingNotifier(i->iSysSocket, (unsigned char *) m_szPacketBuffer, iLength, i->pUserData);

								i->ulLastAction = time(0);
							}							

							break;

						case ST_DATAGRAM:
							{
								struct sockaddr_in addrSource;
								socklen_t iSize = sizeof(addrSource);
								int iLength = recvfrom(i->iSysSocket, m_szPacketBuffer, PACKET_BUFFER_SIZE, 0, (struct sockaddr *) &addrSource, &iSize);

                                if(iLength <= 0)
                                    closeSocket(i->iSysSocket);
                                else
									i->fnDatagramNotifier(i->iSysSocket, (unsigned char *) m_szPacketBuffer, iLength, i->pUserData, addrSource.sin_addr.s_addr, ntohs(addrSource.sin_port));

								i->ulLastAction = time(0);
							}

							break;

						case ST_SERVERLISTEN:
							{
								struct sockaddr_in addrRemote;
								socklen_t iSize = sizeof(addrRemote);
								int iNew = accept(i->iSysSocket, (struct sockaddr *) &addrRemote, &iSize);

								if(iNew > 0)
									addClient(&(* i), iNew, &addrRemote);
                                else
                                    closeSocket(i->iSysSocket);

								i->ulLastAction = time(0);
							}

							break; 

						default:
							m_pLogManager->log(LL_CRITICAL, "Unknown socket type in %s:%i!", __PRETTY_FUNCTION__, __LINE__);
						}
					}
				}
			}

			free(pDescriptors);
		}
	}


	unsigned long NetworkInterface::lookupHost(const char * szHost)
	{ // resolve hostname to ip, blocking (!)
		struct hostent * pHostEnt;

		if(!(pHostEnt = gethostbyname(szHost)))
			return 0;

		return * * ((unsigned long * *) pHostEnt->h_addr_list);
	}


	void NetworkInterface::setKeepAlive(int iHandle, bool bKeepAlive)
	{
		for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i)
		{
			if(i->iSysSocket == iHandle)
			{
				i->bKeepAlive = bKeepAlive;

				return;
			}
		}
	}

	void NetworkInterface::setTimeout(int iHandle, unsigned long ulTimeout)
	{
		for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i)
		{
			if(i->iSysSocket == iHandle)
			{
				i->ulTimeout = ulTimeout;

				return;
			}
		}
	}

	void NetworkInterface::timeoutNoise(int iHandle)
	{
		for(std::list<socket_t>::iterator i = m_lsSockets.begin(); i != m_lsSockets.end(); ++i)
		{
			if(i->iSysSocket == iHandle)
			{
				i->ulLastAction = time(0);

				return;
			}
		}
	}
}
