/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * submit-file.cpp - simple submission module, that copies data to a different location on the hdd and optinally changes owner
 *
 * 2004-01-27	wrote this one
 */

#include <mwcollectd.h>

#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string>
#include <time.h>
#include <errno.h>
#include <unistd.h>


using namespace mwcollectd;

// ============================================================

class SubmitFileModule : public Module, public SubmissionModule
{
public:
	SubmitFileModule(MalwareCollector * pCollector) { m_pCollector = pCollector; }
	virtual ~SubmitFileModule() { }

	virtual void setLogManager(LogManager * pLogManager) { m_pLogManager = pLogManager; }
	virtual void setConfiguration(Configuration * pConfig, const char * szPath) { m_pConfig = pConfig; m_szConfigPath = strdup(szPath); }

	virtual void start();
	virtual void stop();

	virtual module_type_t moduleType() { return MT_SUBMISSION_AGENT; }

	virtual void submitFile(const char * szFilePath);
	virtual void submitData(const unsigned char * szData, unsigned int nLength);

protected:
	MalwareCollector * m_pCollector;
	LogManager * m_pLogManager;

	Configuration * m_pConfig;
	char * m_szConfigPath;
	gid_t m_gid;
	uid_t m_uid;
};

// ============================================================

extern "C"
{
	Module * __attribute__ ((dllexport)) CreateInstance(MalwareCollector * pCollector)
	{
		return (Module *) (new SubmitFileModule(pCollector));
	}
}

// ============================================================

void SubmitFileModule::start()
{
	umask(S_IWGRP | S_IWOTH | S_IROTH | S_IXUSR | S_IXGRP | S_IXOTH);	

	{
		char * szPath, * szOwner, * szGroup;

		asprintf(&szPath, "%s:chown", m_szConfigPath);
		szOwner = strdup(m_pConfig->getString(szPath, "nobody:nogroup"));
		free(szPath);

		szGroup = strchr(szOwner, ':');
		* szGroup = 0;
		++szGroup;

		{
			struct group * pGroup;
			struct passwd * pUser;

			m_uid = 0;
			m_gid = 0;

			if(!(pGroup = getgrnam(szGroup)))
				m_pLogManager->log(LL_CRITICAL, "Could not resolve groupname %s: %i %s", szGroup, errno, strerror(errno));
			else
			{
				if(!(pUser = getpwnam(szOwner)))
					m_pLogManager->log(LL_CRITICAL, "Could not resolve username %s: %i %s", szOwner, errno, strerror(errno));
				else
				{
					m_uid = pUser->pw_uid;
					m_gid = pGroup->gr_gid;
				}
			}
		}

		free(szOwner);
	}

	m_pCollector->getSubmissionDispatcher()->addSubmissionModule(this);
}

void SubmitFileModule::stop()
{
	m_pCollector->getSubmissionDispatcher()->removeSubmissionModule(this);

	free(m_szConfigPath);
}

void SubmitFileModule::submitFile(const char * szFilePath)
{
	const char * szDestination;
	std::string sRealPath;

	{
		{
			char * szPath;

			asprintf(&szPath, "%s:destination", m_szConfigPath);
			szDestination = m_pConfig->getString(szPath, "/var/mwcollect/");
			free(szPath);
		}

		{
			sRealPath = std::string(szDestination);

			if(sRealPath[sRealPath.length() - 1] != '/')
				sRealPath.push_back('/');

			if(strrchr(szFilePath, '/'))
				sRealPath += std::string(strrchr(szFilePath, '/') + 1);
			else
				sRealPath += std::string(szFilePath);
		}
	}

	{
		FILE * a, * b;
		char szBuffer[4096];
		int iLen;

		if(!(a = fopen(szFilePath, "rb")) || !(b = fopen(sRealPath.c_str(), "wb")))
			return;

		while((iLen = fread(szBuffer, 1, sizeof(szBuffer), a)))
			if(!fwrite(szBuffer, 1, iLen, b))
				break;

		fclose(b);
		fclose(a);
	}

	if(chown(sRealPath.c_str(), m_uid, m_gid) < 0)
		m_pLogManager->log(LL_DEBUG, "Chown of %s to %u:%u failed: %i %s.", sRealPath.c_str(), (unsigned int) m_uid, (unsigned int) m_gid, errno, strerror(errno));

	m_pLogManager->log(LL_DEBUG, "submitFile \"%s\" => \"%s\" successful.", szFilePath, sRealPath.c_str());
}

void SubmitFileModule::submitData(const unsigned char * szData, unsigned int nLength)
{
	const char * szDestination;
	std::string sRealPath;

	{
		{
			char * szPath;

			asprintf(&szPath, "%s:destination", m_szConfigPath);
			szDestination = m_pConfig->getString(szPath, "/var/mwcollect/");
			free(szPath);
		}

		{
			char * szSuffix;

			asprintf(&szSuffix, "%u.%08x", time(0), rand());

			sRealPath = std::string(szDestination);

			if(sRealPath[sRealPath.length() - 1] != '/')
				sRealPath.push_back('/');

			sRealPath += std::string("mwc-datasubm.");
			sRealPath += std::string(szSuffix);

			free(szSuffix);
		}
	}

	{
		FILE * pFile = fopen(sRealPath.c_str(), "wb");

		if(!pFile || !fwrite(szData, 1, nLength, pFile))
		{
			m_pLogManager->log(LL_CRITICAL, "Could not write data to file \"%s\"!", sRealPath.c_str());

			return;
		}

		fclose(pFile);
	}

	if(chown(sRealPath.c_str(), m_uid, m_gid) < 0)
		m_pLogManager->log(LL_DEBUG, "Chown of %s to %u:%u failed: %i %s.", sRealPath.c_str(), (unsigned int) m_uid, (unsigned int) m_gid, errno, strerror(errno));
}
