/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * bagle.cpp - bagle vulnerability, one dirty file
 *
 * 2005-01-11	wrote this one
 * 2005-01-16	rewrite since it was crap..
 */

/* we parse the auth signatures from the config file
 * so this one can be easily extended by config to accept other
 * "$send_magic && $send_url"-virii
 */

#include <mwcollectd.h>
#include <list>
#include <string>

using namespace mwcollectd;

typedef struct
{
	unsigned char * pucSignature; // parsed signature
	unsigned int nLength; // length of parsed signature
} signature_t;

typedef struct
{
	bool fMagicSent;
	bool fMatch;
	bool fDone;

	NetworkInterface * pNetworking;
	LogManager * pLogManager;
	FetchDispatcher * pFetchDispatcher;

	std::list<signature_t> * plsSignatures;
} bagle_state_t;

class BagleModule : public Module
{
public:
	BagleModule(MalwareCollector * pCollector);
	virtual ~BagleModule() { }

	virtual void setLogManager(LogManager * pLogManager);
	virtual void setConfiguration(Configuration * pConfig, const char * szPath);

	virtual void start();
	virtual void stop();

	virtual module_type_t moduleType() { return MT_COLLECTOR; }

protected:
	LogManager * m_pLogManager;
	MalwareCollector * m_pCollector;

	Configuration * m_pConfig;
	char * m_szConfigPath;

	int m_iServerSocket;

	std::list<signature_t> m_lsSignatures;
};

void * BagleConnectionNotifier(int iHandle, unsigned long ulSourceAddress, unsigned short usSourcePort, unsigned short usLocalPort, void * pUser);
void BagleIncomingNotifier(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam);
void BagleTerminationNotifier(bool bServer, void * pParam);



extern "C"
{
  mwcollectd::Module * __attribute__ ((dllexport)) CreateInstance(mwcollectd::MalwareCollector * pCollector)
  {
    return (new BagleModule(pCollector));
  }
}



BagleModule::BagleModule(MalwareCollector * pCollector)
{
	m_pCollector = pCollector;

	m_szConfigPath = 0;
}

void BagleModule::setLogManager(LogManager * pLogManager)
{
	m_pLogManager = pLogManager;
}

void BagleModule::setConfiguration(Configuration * pConfig, const char * szPath)
{
	m_pConfig = pConfig;
	m_szConfigPath = strdup(szPath);
}


void BagleModule::start()
{
	{ // parse signatures
		signature_t signCurrent;
		char * szConfPath;;
		std::list<const char *> * pConfSignatures;

		asprintf(&szConfPath, "%s:signatures", m_szConfigPath);
		pConfSignatures = m_pConfig->getChildren(szConfPath);
		free(szConfPath);

		if(pConfSignatures)
		{
			for(std::list<const char *>::iterator i = pConfSignatures->begin(); i != pConfSignatures->end(); ++i)
			{
				char szCopy[3];
				const char * szParse;

				szCopy[2] = 0;

				asprintf(&szConfPath, "%s:signatures:%s", m_szConfigPath, * i);
				szParse = m_pConfig->getString(szConfPath, 0);
				free(szConfPath);

				if(strlen(szParse) % 2)
				{
					m_pLogManager->log(LL_CRITICAL, "Signature %s has uneven length, dropping!", * i);

					continue;
				}

				signCurrent.nLength = strlen(szParse) / 2;
				signCurrent.pucSignature = (unsigned char *) malloc(signCurrent.nLength);

				for(unsigned int j = 0; j < signCurrent.nLength; ++j)
				{
					szCopy[0] = szParse[j * 2];
					szCopy[1] = szParse[j * 2 + 1];

					signCurrent.pucSignature[j] = (unsigned char) (strtoul(szCopy, 0, 16) & 0xFF);
				}

				m_lsSignatures.push_back(signCurrent);

				m_pLogManager->log(LL_SPAM, "Added signature %s.", * i);
			}
		}
		else
			m_pLogManager->log(LL_CRITICAL, "No Bagle signatures defined in config file!");
	}

	{ // create server socket
		long lPort;

		{
			char * szPath;

			asprintf(&szPath, "%s:port", m_szConfigPath);
			lPort = m_pConfig->getLong(szPath, 2745);
			free(szPath);
		}

		{
			bagle_state_t * pState = new bagle_state_t;

			pState->fMagicSent = true;
			pState->fDone = false;
			pState->pNetworking = m_pCollector->getNetworkInterface();
			pState->pLogManager = m_pLogManager;
			pState->pFetchDispatcher = m_pCollector->getFetchDispatcher();
			pState->plsSignatures = &m_lsSignatures;

			m_iServerSocket = m_pCollector->getNetworkInterface()->createSocket((unsigned short) lPort, BagleConnectionNotifier, BagleIncomingNotifier, BagleTerminationNotifier, pState);
		}
	}
}

void BagleModule::stop()
{
	m_pCollector->getNetworkInterface()->closeSocket(m_iServerSocket);

	{ // free parsed signature pointers
		for(std::list<signature_t>::iterator i = m_lsSignatures.begin(); i != m_lsSignatures.end(); ++i)
			free(i->pucSignature);
	}
}

void * BagleConnectionNotifier(int iHandle, unsigned long ulSourceAddress, unsigned short usSourcePort, unsigned short usLocalPort, void * pUser)
{
	bagle_state_t * pState = new bagle_state_t;
	in_addr addr;

	memcpy(pState, pUser, sizeof(bagle_state_t));
	pState->fMagicSent = false;
	pState->fMatch = false;

	addr.s_addr = ulSourceAddress;
	pState->pLogManager->log(LL_DEBUG, "Bagle connection from %s:%u (to :%u).", inet_ntoa(addr), (unsigned int) usSourcePort, (unsigned int) usLocalPort);

	return pState;
}

void BagleIncomingNotifier(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam)
{
	bagle_state_t * pState = (bagle_state_t *) pParam;

	if(pState->fDone)
		return;

	if(pState->fMagicSent /* && pState->fMatch */)
	{
		char * szUrl = (char *) malloc(nPacketLength + 1);
		char * szWalk;

		memcpy(szUrl, pucPacketData, nPacketLength);
		szUrl[nPacketLength] = 0;

		for(szWalk = szUrl; * szWalk; ++szWalk)
			if(* szWalk == '\r' || * szWalk == '\n')
			{
				* szWalk = 0;
				break;
			}

		pState->pLogManager->log(LL_INFO, "Successful bagle session, fetch \"%s\".", szUrl);
		pState->pFetchDispatcher->fetchFile(szUrl);

		free(szUrl);

		pState->fDone = true;
		pState->pNetworking->closeSocket(iHandle);		
	}
	else
	{
		pState->fMatch = false;

		for(std::list<signature_t>::iterator i = pState->plsSignatures->begin(); i != pState->plsSignatures->end(); ++i)
			if(i->nLength == nPacketLength && !memcmp(i->pucSignature, pucPacketData, nPacketLength))
				pState->fMatch = true;

		pState->fMagicSent = true;

		if(!pState->fMatch)
		{
			char * szHexdump = (char *) malloc(nPacketLength * 2 + 1);

			for(unsigned int i = 0; i < nPacketLength; ++i)
			{
				szHexdump[i * 2 + 0] = ((pucPacketData[i] >> 4) < 10 ? (pucPacketData[i] >> 4) + '0' : (pucPacketData[i] >> 4) + ('A' - 10));
				szHexdump[i * 2 + 1] = ((pucPacketData[i] & 0xF) < 10 ? (pucPacketData[i] & 0xF) + '0' : (pucPacketData[i] & 0xF) + ('A' - 10));
			}

			szHexdump[nPacketLength * 2] = 0;

			pState->pLogManager->log(LL_INFO, "Bagle session with invalid auth string: %s", szHexdump);
			free(szHexdump);
		
			pState->pNetworking->sendData(iHandle, (unsigned char *) "send your shit...", 17);
		}
		else
		{
			pState->pLogManager->log(LL_DEBUG, "Bagle authentication successful.");			
		}

		pState->pNetworking->sendData(iHandle, (unsigned char *) "send your shit...", 17);
	}
}

void BagleTerminationNotifier(bool bServer, void * pParam)
{
	if(!bServer)
		delete (bagle_state_t *) pParam;
}
