/* mwcollect2 (c) 2004-2005 Georg Wicherski
 * this code is licensed under the mwcollect2 license, the the file LICENSE for details
 * distribution of this code without written permission strictly prohibited
 *
 * dispatch.h - definitions of the dispatcher classes
 *
 * 2004-12-28	created, added FetchDispatcher && SubmissionDispatcher
 * 2004-01-19	added shellcode dispatcher
 */

#ifndef __MWCOLLECTD_DISPATCH_H
#define __MWCOLLECTD_DISPATCH_H

#include <list>
#include <map>

namespace mwcollectd
{
	typedef struct
	{
		unsigned long m_ulAddress;
		unsigned long m_ulMask;
	} ip_range_t;

	class SubmissionModule // submission module interface
	{
	public:
		virtual void submitFile(const char * szFilePath) = 0; // if the malware has already been put into a temporary file, DELETION up to you
		virtual void submitData(const unsigned char * szData, unsigned int nLength) = 0; // submit in memory malware
	};

	class SubmissionDispatcher // dispatches submission info, if a file has been leeched, some ppl could want to pack into a sql db etc..
	{
	public:
		SubmissionDispatcher(LogManager * pLogManager, Configuration * pConfiguration);
		virtual ~ SubmissionDispatcher();

		virtual void addSubmissionModule(SubmissionModule * pSubm);
		virtual void removeSubmissionModule(SubmissionModule * pSubm);

		virtual void submitFile(const char * szFilePath); // if the malware has already been put into a temporary file, DELETION up to you
		virtual void submitData(const unsigned char * szData, unsigned int nLength); // submit in memory malware

	protected:
		std::list<SubmissionModule *> m_lpModules;

		LogManager * m_pLogManager;
		Configuration * m_pConfig;
	};

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	struct url_eq
	{
		bool operator()(const char * a, const char * b) { return strcmp(a, b) < 0; }
	};

	class ProtocolHandler // interface to implement if you want to fetch files blafoo
	{
	public:
		virtual void fetchFile(const char * szUrl) = 0; // put the selected file into the fetch queue

		virtual bool supportProtocol(const char * szProtocol) = 0;
	};

	class FetchDispatcher // coordinates url fetch request
	{
	public:
		FetchDispatcher(SubmissionDispatcher * pSubmissionDispatcher, LogManager * pLogManager, Configuration * pConfiguration);
		virtual ~FetchDispatcher();

		virtual void addProtocolHandler(ProtocolHandler * pHandler);
		virtual void removeProtocolHandler(ProtocolHandler * pHandler);

		virtual void fetchFile(const char * szUrl); // dispatch the fetch request (gets submitted afterwards)

        static bool isLocalAddress(unsigned long ulAddress);

	protected:
		std::list<ProtocolHandler *> m_lpHandlers;
		std::map<const char *, unsigned long, url_eq> m_mUniq;

		SubmissionDispatcher * m_pSubmDispatcher;

		LogManager * m_pLogManager;
		Configuration * m_pConfig;

		static ip_range_t m_irLocalRanges[];
	};

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/* the two dispatchers work in a slight different fashion,
	 * SubmissionDispatcher dispatches one of the two possible request to all registered modules, while
	 * FetchDispatcher dispatches the request only to one matching protocol handler
	 */

	class ShellcodeParser
	{
	public:
		virtual bool parseShellcode(unsigned char * pucShellcode, unsigned int nLength, unsigned long ulSource) = 0;
	};

	class ShellcodeDispatcher
	{
	public:
		ShellcodeDispatcher(LogManager * pLogManager);
		virtual ~ ShellcodeDispatcher();

		virtual void addShellcodeParser(ShellcodeParser * p);
		virtual void removeShellcodeParser(ShellcodeParser * p);

		virtual bool parseShellcode(unsigned char * pucShellcode, unsigned int nLength, unsigned long ulSourceAddress);

	protected:
		std::list<ShellcodeParser *> m_lpParsers;

		LogManager * m_pLogManager;
	};
}

#endif // __MWCOLLECTD_DISPATCH_H
