/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * session.cpp - the substuff that handles the packets, one session per connect...
 *
 * 2005-01-20	added (just became 17 years old, yay...)
 */

#include "vuln-win32rpc.h"
#include "dcom-requests.h"
#include "dcom2-requests.h"
#include "lsass-requests.h"

RpcSession::RpcSession(MalwareCollector * pCollector, LogManager * pLogManager)
{
	m_pCollector = pCollector;
	m_pLogManager = pLogManager;
}

RpcSession::~RpcSession()
{
}

void RpcSession::etablishedConnection(unsigned long ulSource, unsigned short usSourcePort, unsigned short usLocalPort)
{
	struct in_addr addr;
	addr.s_addr = ulSource;

	m_dsState = RPCS_CONNECTED;
	m_ulSource = ulSource;
	m_usLocalPort = usLocalPort;

	m_pLogManager->log(LL_SPAM, "Incoming DCOM connection from %s:%u (to :%u).", inet_ntoa(addr), usSourcePort, usLocalPort, this);
}

unsigned int safe_strlen(const char * p, unsigned int nmax)
{
	unsigned int i = 0;

	while(* (p++) && (i++) < nmax);

	return i;
}

void RpcSession::incomingData(int iHandle, unsigned char * pucData, unsigned int nLength)
{
	switch(m_dsState)
	{
	case RPCS_CONNECTED:
		m_pLogManager->log(LL_SPAM, "Initial RPC BindString received, responding with crap.");

		{
			unsigned char szBuffer[256];

			for(unsigned int i = 0; i < sizeof(szBuffer); ++i)
				szBuffer[i] = rand() % 0xFF;

			m_pCollector->getNetworkInterface()->sendData(iHandle, szBuffer, sizeof(szBuffer));

			if(nLength >= sizeof(dcom_bindstr) - 1 && memcmp(dcom_bindstr, pucData, sizeof(dcom_bindstr) - 1) == 0)
			{
				m_pLogManager->log(LL_SPAM, "Valid classic DCOM BindString.");

				m_dsState = RPCS_GOT_DCOM_BINDSTR;
			}
			else if(nLength >= sizeof(dcom2_bindstr) - 1 && memcmp(dcom2_bindstr, pucData, sizeof(dcom2_bindstr) - 1) == 0)
			{
				m_pLogManager->log(LL_SPAM, "Valid DCOM2 BindString.");

				m_dsState = RPCS_GOT_DCOM_BINDSTR;
			}
			else if(nLength >= sizeof(lsass_req1) - 1 && memcmp(lsass_req1, pucData, sizeof(lsass_req1) - 1) == 0)
			{
				m_pLogManager->log(LL_SPAM, "Valid LSASS Request #1.");

				m_dsState = RPCS_GOT_LSASS_STAGE1;
			}
			else
			{
				{
					char * szHexdump = (char *) malloc(nLength * 2 + 1);

					for(unsigned int i = 0; i < nLength; ++i)
					{
						szHexdump[i * 2 + 0] = ((pucData[i] >> 4) < 10 ? (pucData[i] >> 4) + '0' : (pucData[i] >> 4) + ('A' - 10));
						szHexdump[i * 2 + 1] = ((pucData[i] & 0xF) < 10 ? (pucData[i] & 0xF) + '0' : (pucData[i] & 0xF) + ('A' - 10));
					}

					szHexdump[nLength * 2] = 0;

					m_pLogManager->log(LL_INFO, "Invalid RPC BindString to :%u from %s, Hexdump: %s", m_usLocalPort, inet_ntoa(* ((in_addr *) &m_ulSource)), szHexdump);
				}

				m_dsState = RPCS_OBSOLETE;
				m_pCollector->getNetworkInterface()->closeSocket(iHandle);				
		
				return;
			}
		}		

		break;

	case RPCS_GOT_DCOM_BINDSTR:
		{
			if(nLength < 0x10 + sizeof(unsigned long))
			{
				m_pLogManager->log(LL_SPAM, "Got undersized DCOM Packet! %u", this);

				return;
			}

			unsigned long ulShellcodeLength = * ((unsigned long *) (pucData + 0x10)) - 964;

			if(nLength < 880 + ulShellcodeLength)
			{
				m_pLogManager->log(LL_SPAM, "Got undersized DCOM Packet: %u bytes shellcode would require %u bytes Packet, but is %u bytes long. %u", ulShellcodeLength, ulShellcodeLength + 880, nLength, this);

				return;
			}

			m_pLogManager->log(LL_DEBUG, "DCOM Shellcode starts at byte 0x%04X and is 0x%04X bytes long.", 880, ulShellcodeLength);

			m_pCollector->getShellcodeDispatcher()->parseShellcode(pucData + 880, ulShellcodeLength, m_ulSource);

			m_dsState = RPCS_OBSOLETE;
			m_pCollector->getNetworkInterface()->closeSocket(iHandle);
		}		

		break;

	case RPCS_GOT_LSASS_STAGE1:
		if(!(nLength >= sizeof(lsass_req2) - 1 && memcmp(lsass_req2, pucData, sizeof(lsass_req2)) == 0))
			m_pLogManager->log(LL_SPAM, "Invalid Stage II LSASS Packet.");

		{
			unsigned char szBuffer[256];

			for(unsigned int i = 0; i < sizeof(szBuffer); ++i)
				szBuffer[i] = rand() % 0xFF;

			m_pCollector->getNetworkInterface()->sendData(iHandle, szBuffer, sizeof(szBuffer));

			m_dsState = RPCS_GOT_LSASS_STAGE2;
		}
			

		break;

	case RPCS_GOT_LSASS_STAGE2:
		if(!(nLength >= sizeof(lsass_req3) - 1 && memcmp(lsass_req3, pucData, sizeof(lsass_req3)) == 0))
			m_pLogManager->log(LL_SPAM, "Invalid Stage III LSASS Packet.");

		{
			unsigned char szBuffer[1024];
			char * szIdentity = "Windows 5.1 ";

			for(unsigned int i = 0; i < sizeof(szBuffer); ++i)
				szBuffer[i] = rand() % 0xFF;

			for(unsigned int i = 0; szIdentity[i]; ++i)
				szBuffer[48 + i * 2] = szIdentity[i];

			m_pCollector->getNetworkInterface()->sendData(iHandle, szBuffer, sizeof(szBuffer));

			m_dsState = RPCS_GOT_LSASS_STAGE3;
		}

		break;

	case RPCS_GOT_LSASS_STAGE3:
	case RPCS_GOT_LSASS_STAGE4:
	case RPCS_GOT_LSASS_STAGE5:
		{
			unsigned char szBuffer[256];

			for(unsigned int i = 0; i < sizeof(szBuffer); ++i)
				szBuffer[i] = rand() % 0xFF;

			m_pCollector->getNetworkInterface()->sendData(iHandle, szBuffer, sizeof(szBuffer));

			m_dsState = (rpc_state_t) ((unsigned int) m_dsState + 1);
		}

		break;

	case RPCS_GOT_LSASS_STAGE6:
		{
			if(nLength < 0x20)
				m_pLogManager->log(LL_SPAM, "Don't know how to handle small LSASS packet...");
			else
			{
				if(nLength >= sizeof(lsass_req7) - 1 && memcmp(pucData, lsass_req7, sizeof(lsass_req7) - 1) == 0)
				{
					m_pLogManager->log(LL_DEBUG, "Classic House of Dabus LSASS Stage VII detected.");

					m_pCollector->getShellcodeDispatcher()->parseShellcode(pucData + sizeof(lsass_req7) - 1, safe_strlen((char *) pucData + sizeof(lsass_req7) - 1, nLength - sizeof(lsass_req7) + 1), m_ulSource);
				}
				else if(nLength >= sizeof(lsass_inthewild1_stage7) - 1 && memcmp(pucData, lsass_inthewild1_stage7, sizeof(lsass_inthewild1_stage7) - 1) != 0)
				{
					m_pLogManager->log(LL_DEBUG, "In-The-Wild LSASS stage VII type I");

					m_pCollector->getShellcodeDispatcher()->parseShellcode(pucData + 0x9A, nLength - 0x9A, m_ulSource);
				}
				else // no signature but still some tests left dude ;)
				{
					bool bNopslide = false, bBrokenbot = false;
					unsigned long ulOffset;

					{
						unsigned long i;

						for(i = 0; i < nLength && pucData[i] == 0x31; ++i);

						if(pucData[i] == 0x00 && nLength - i < 3)
							bBrokenbot = true;
					}
					
					if(!bBrokenbot)
					{
						unsigned long i;

						for(i = 0; i < nLength && pucData[i] == 0x90; ++i);

						if(i > 0x30 && nLength - i > 0x20)
						{
							ulOffset = i - 0x10;
							bNopslide = true;
						}
					}

					if(bBrokenbot)
					{
						m_pLogManager->log(LL_DEBUG, "Got broken-bot LSASS stage VIII, discarding.");
					}
					else if(bNopslide)
					{
						m_pLogManager->log(LL_DEBUG, "Got generic nopslide LSASS stage VIII, analyzing...");

						m_pCollector->getShellcodeDispatcher()->parseShellcode(pucData + ulOffset, nLength - ulOffset, m_ulSource);
					}
					else
					{
						char * szHexdump = (char *) malloc(nLength * 2 + 1);

						for(unsigned int i = 0; i < nLength; ++i)
						{
							szHexdump[i * 2 + 0] = ((pucData[i] >> 4) < 10 ? (pucData[i] >> 4) + '0' : (pucData[i] >> 4) + ('A' - 10));
							szHexdump[i * 2 + 1] = ((pucData[i] & 0xF) < 10 ? (pucData[i] & 0xF) + '0' : (pucData[i] & 0xF) + ('A' - 10));
						}

						szHexdump[nLength * 2] = 0;

						m_pLogManager->log(LL_DEBUG, "Unknown LSASS Stage VII, trying anyway; Hexdump: %s", szHexdump);
						free(szHexdump);

						m_pCollector->getShellcodeDispatcher()->parseShellcode(pucData + sizeof(lsass_req7) - 1, safe_strlen((char *) pucData + sizeof(lsass_req7) - 1, nLength - sizeof(lsass_req7) + 1), m_ulSource);
					}
				}
			}

		}

		m_dsState = RPCS_OBSOLETE;
		m_pCollector->getNetworkInterface()->closeSocket(iHandle);		

		break;

	case RPCS_OBSOLETE:
		break;

	default:
		m_pLogManager->log(LL_CRITICAL, "Unknown DCOM State (%i)!", m_dsState);
	}
}

void RpcSession::connectionTerminated()
{
}
